
#include <string.h>
#include <libnautilus-extension/nautilus-info-provider.h>
#include <unistd.h>
#include <stdio.h>
#include <gconf/gconf.h>
#include <eel/eel-preferences.h>
#include <glib-2.0/glib/gthread.h>
#include <stdlib.h>
#include "nautilus-metadata.h"
#include "totem-mime-types.h"

/* Nautilus extension interface (standard stuff, don't look at these) */
void nautilus_module_initialize (GTypeModule  *module);
void nautilus_module_list_types (const GType **types, int *num_types);
void nautilus_module_shutdown   (void);

/* 
 * Nautilus extension related stuff
 * look only at:
 * metadata_info_update_file_info()   <- executed for every file/dir in the current nautilus window
 * fill_metadata_async()    <- executed in a thread created by the proevious function
 */
static GType tpp_type = 0;
typedef struct {
	GClosure *update_complete;
	NautilusInfoProvider *provider;
	NautilusFileInfo *file;
	gboolean cancelled;
} UpdateHandle;
static void metadata_info_provider_iface_init (NautilusInfoProviderIface *iface);
static void metadata_info_cancel_update(NautilusInfoProvider *provider, NautilusOperationHandle *handle);
static void metadata_info_plugin_register_type(GTypeModule  *module);
static NautilusOperationResult metadata_info_update_file_info(NautilusInfoProvider *provider, NautilusFileInfo *file, GClosure *update_complete, NautilusOperationHandle **handle);
static gpointer fill_metadata_async(gpointer update_handle);

/* 
 * collect Nautilus preferences to execute the functions to collect metadata
 * only if Nautilus is displaying the corresponding attributes in the icon/list view
 */
static gboolean icon_view_is_interested = FALSE;
static gboolean list_view_is_interested = FALSE;
static guint gconf_icon_view_uid;
static guint gconf_list_view_uid;
static void gconf_key_changed (GConfClient* conf, guint cnxn_id, GConfEntry *entry, gpointer user_data);

/* thread limiting stuff */
#define MAX_THREADS 5
static guint executing_threads = 0;
static guint cond_processes = 0;
static GMutex *mutex;
static GCond *cond;

/* collect metadata stuff */
#define BUF_DIM 1000
static void datalist_free(gpointer datalist);
GData** collect_metadata(const char* uri);
static void tags_foreach(GQuark key_id, gpointer data, gpointer user_data);

/*
 * this function is called to fill NautilusFileInfo with the collected 
 * attribute contained in a GData object
 */
static void tags_foreach(GQuark key_id, gpointer data, gpointer user_data)
{
    NautilusFileInfo* file = (NautilusFileInfo*) user_data;
    NautilusFileInfoIface* iface = NAUTILUS_FILE_INFO_GET_IFACE(file);
    iface->add_string_attribute(file, g_quark_to_string(key_id), (char*) data);
}

static void datalist_free(gpointer datalist_p)
{
    GData** datalist = (GData**) datalist_p;
    g_datalist_clear(datalist);   
    g_free(datalist);
}

/*
 * this function is automatically called whenever a registered GConf key changes.
 * the registered keys are:
 * NAUTILUS_GCONF_ICON_VIEW_KEY "/apps/nautilus/icon_view/captions"
 * NAUTILUS_GCONF_LIST_VIEW_KEY "/apps/nautilus/list_view/default_visible_columns"
 */
static void gconf_key_changed (GConfClient* conf, guint cnxn_id, GConfEntry *entry, gpointer user_data)
{
    const char* key = gconf_entry_get_key(entry);
    gboolean used = FALSE;
    
    // NOT WORKING CODE
    //GConfValue* value = gconf_entry_get_value(entry);
    // this does not work, the GSList is filled with empty strings: {'', '\0'}
    //GSList *values = gconf_value_get_list(value), *values_first=values;
    
    GSList *values = eel_gconf_get_string_list(key);
    char* attr;
    int i;
    // check if any of the string in the list, is a metadata attribute (contained in media_attributes_provided[])
    do
    {
        attr = (char*) values->data;
        for (i=0; i<G_N_ELEMENTS(media_attributes_provided); i++)
        {
            if (strcmp(attr, media_attributes_provided[i]) == 0)
            {
                used = TRUE;
                break;
            }
        }
        
        values = g_slist_next(values);
    }
    while(used==FALSE && values!=NULL);
    
    *(gboolean*)user_data = used;
}

/*
 * this is the function that nautilus calls when it lists a directory
 * this function blocks the metadata gathering in the following cases:
 * - the file has a wrong mime type
 * - nautilus views don't have any multimedia metadata enabled
 */
static NautilusOperationResult metadata_info_update_file_info(NautilusInfoProvider *provider, NautilusFileInfo *file, GClosure *update_complete, NautilusOperationHandle **handle)
{
        NautilusFileInfoIface* iface = NAUTILUS_FILE_INFO_GET_IFACE(file);
        
        // return if nautilus doesn't use the attributes
        if (!icon_view_is_interested && !list_view_is_interested)
        {
            return NAUTILUS_OPERATION_COMPLETE;
        }
        
        // return if this is a directory
        if (iface->is_directory(file))
        {
            return NAUTILUS_OPERATION_COMPLETE;
        }
        
        // return if this file has a wrong mime_type
        gboolean audio_video = FALSE;
        int i=G_N_ELEMENTS(mime_types) - 1;

        while (audio_video == FALSE && i>=0)
        {
            if (iface->is_mime_type(file, mime_types[i]))
            {
                audio_video = TRUE;
            }
            i--;
        }

        if (audio_video == FALSE)
        {
            return NAUTILUS_OPERATION_COMPLETE;
        }
        
        /* TRY TO GET CACHED RESULTS */
	
	GData **tags;
        
        tags = g_object_get_data (G_OBJECT (file), FILE_HAS_METADATA);
	if (tags != NULL)
	{
                /* FILL NAUTILUS ATTRIBUTES USING CACHED RESULTS */
                g_datalist_foreach(tags, tags_foreach, file);
		return NAUTILUS_OPERATION_COMPLETE;
	}
	else
	{
                /* NO CACHED RESTULTS FOUND */
		UpdateHandle *update_handle = g_new (UpdateHandle, 1);
		update_handle->update_complete = g_closure_ref (update_complete);
		update_handle->provider = provider;
		update_handle->file = g_object_ref (file);
                update_handle->cancelled = FALSE;
		*handle = (NautilusOperationHandle*) update_handle;
		
		GError *error = NULL;
                /* ASYNCHRONOUSLY FILL NAUTILUS ATTRIBUTES USING FUNCTION async(update_handle) */
		g_thread_create(fill_metadata_async, update_handle, FALSE, &error);
		if (error != NULL)
		{
			g_error_free(error);
			return NAUTILUS_OPERATION_FAILED;
		}
	
		return NAUTILUS_OPERATION_IN_PROGRESS;
	}
}

/*
 * called by metadata_info_update_file_info
 * this function is executed in a thread, only if the file is audio/video and if
 * the extension considers gathering metadata useful in the current environment
 */
static gpointer fill_metadata_async(gpointer update_handle)
{    
	// this code is executed asynchronously
	UpdateHandle* handle = (UpdateHandle*) update_handle;
        NautilusFileInfo *file = handle->file;
        NautilusFileInfoIface *file_iface = NAUTILUS_FILE_INFO_GET_IFACE(file);        
        
        // temporary stop the thread if MAX_THREADS is reached
        g_mutex_lock(mutex);
        if (executing_threads >= MAX_THREADS)
        {
            cond_processes++;
            g_cond_wait(cond, mutex);
            cond_processes--;
        }
        executing_threads++;
        g_mutex_unlock(mutex);
        
        GData** tags = NULL;
        if (!handle->cancelled)
        {
            char* uri = file_iface->get_uri(file);
            tags = collect_metadata(uri);
            g_free(uri);

            /* FILL NAUTILUS ATTRIBUTES */
            g_datalist_foreach(tags, tags_foreach, handle->file);
            
            /* INFORM NAUTILUS WE FINISHED FILLING ATTRIBUTES */
            nautilus_info_provider_update_complete_invoke(handle->update_complete, handle->provider, (NautilusOperationHandle*)handle, NAUTILUS_OPERATION_COMPLETE);
        }

        /* CACHING RESULTS (only cached for a bunch of seconds) */ 
        if (tags != NULL)
            g_object_set_data_full (G_OBJECT (handle->file), FILE_HAS_METADATA, tags, datalist_free);

	g_closure_unref (handle->update_complete);
	g_object_unref (handle->file);
	g_free (handle);
        
        // release other threads that were waiting
        sleep(1);
        g_mutex_lock(mutex);
        executing_threads--;
        if (cond_processes > 0)
            g_cond_signal(cond);
        g_mutex_unlock(mutex);
}

static void metadata_info_provider_iface_init (NautilusInfoProviderIface *iface)
{
	iface->cancel_update = metadata_info_cancel_update;
	iface->update_file_info = metadata_info_update_file_info;
}

static void metadata_info_cancel_update(NautilusInfoProvider  *provider, NautilusOperationHandle  *handle)
{
        UpdateHandle *update_handle = (UpdateHandle*)handle;     
        update_handle->cancelled = TRUE;
}

static void metadata_info_plugin_register_type(GTypeModule  *module)
{	
	const GTypeInfo info = {
		sizeof (GObjectClass),
		(GBaseInitFunc) NULL,
		(GBaseFinalizeFunc) NULL,
		(GClassInitFunc) NULL,
		NULL,
		NULL,
		sizeof (GObject),
		0,
		(GInstanceInitFunc) NULL
	};
	const GInterfaceInfo metadata_info_provider_iface_info = {
		(GInterfaceInitFunc)metadata_info_provider_iface_init,
		NULL,
		NULL
	};

	tpp_type = g_type_module_register_type (module, G_TYPE_OBJECT,
			"MetadataInfoPlugin",
			&info, 0);
	g_type_module_add_interface (module,
			tpp_type,
			NAUTILUS_TYPE_INFO_PROVIDER,
			&metadata_info_provider_iface_info);
}

void nautilus_module_initialize (GTypeModule  *module)
{    
	metadata_info_plugin_register_type(module);
        
        // with this step whenever the gconf key changes the function gconf_key_changed is called
        // so the extension can understand if nautilus shows metadata somewhere
        // currently only in list view or icon view
        gconf_icon_view_uid = eel_gconf_notification_add(NAUTILUS_GCONF_ICON_VIEW_KEY, gconf_key_changed, &icon_view_is_interested);
        gconf_list_view_uid = eel_gconf_notification_add(NAUTILUS_GCONF_LIST_VIEW_KEY, gconf_key_changed, &list_view_is_interested);
        
        // this step is required for initialization
        GConfClient* client = gconf_client_get_default();
        GConfEntry* entry = gconf_client_get_entry(client, NAUTILUS_GCONF_ICON_VIEW_KEY, NULL, TRUE, NULL);
        gconf_key_changed(client, 0, entry, &icon_view_is_interested);
       
        entry = gconf_client_get_entry(client, NAUTILUS_GCONF_LIST_VIEW_KEY, NULL, TRUE, NULL);
        gconf_key_changed(client, 0, entry, &list_view_is_interested);
        
        // initializing thread sync objects
        mutex = g_mutex_new();
        cond = g_cond_new();
}

void nautilus_module_shutdown   (void)
{
    eel_gconf_notification_remove(gconf_icon_view_uid);
    eel_gconf_notification_remove(gconf_list_view_uid);
    g_mutex_free(mutex);
    g_cond_free(cond);
}

void nautilus_module_list_types (const GType **types, int *num_types)
{
	static GType type_list[1];

	type_list[0] = tpp_type;
	*types = type_list;
	*num_types = G_N_ELEMENTS (type_list);
}

/*
 * returns a GData containing all the multimedia tags.
 * it executes the process totem-video-indexer and parses the output
 * it uses fork() exec() pipe() read()
 */
GData** collect_metadata(const char* uri)
{

    int channel[2];
    
    if (pipe(channel)<0)
        return NULL;

    int pid = fork();
    if (pid == 0)
    {
        // child
        close(channel[0]);
        // close stdout and replace with pipe
        dup2(channel[1], STDOUT_FILENO);
        execlp("totem-video-indexer", "totem-video-indexer", uri, NULL);
        // if something goes wrong
        close(channel[1]);
        exit(EXIT_FAILURE);
    }
    else
    {
        // father
        
        // close pipe write channel
        close(channel[1]);
        
        // buffers to read from the pipe, index to the buffer array
        char buffer[BUF_DIM], *buffer_copy;
        char char_buffer;
        int k=0;
        
        // create the datalist which contains the <tag_names, tag_values>
        GData **tags = g_new(GData* ,1);
        g_datalist_init(tags);
        
        // the resolution is composed so it's better to view (1024 x 768)
        // variables to cache the resolution parts x and y
        char *res_x=NULL, *res_y=NULL;
        
        // start to read from the pipe, one character at a time
        while (read(channel[0], &char_buffer, sizeof(char)) > 0)
        {
            // insert the read character into the buffer
            if (k < BUF_DIM - 1)
            {
                buffer[k] = char_buffer;
                k++;
            }
            
            // = means that we finished to read the tag_name
            if (char_buffer == '=')
            {
                buffer[k-1] = '\0';
                // cache the tag_name
                buffer_copy = g_strdup(buffer);
                k=0;
            }
            
            // \n means that we finished to read the tag_value
            else if (char_buffer == '\n')
            {
                buffer[k-1] = '\0';
                
                // video X resolution, cache the result instead of add to the datalist
                if (strcmp(buffer_copy, MEDIA_VIDEO_RES_X) == 0)
                {
                    res_x = g_strdup(buffer);
                }
                // video Y resolution, cache the result instead of add to the datalist
                else if (strcmp(buffer_copy, MEDIA_VIDEO_RES_Y) == 0)
                {
                    res_y = g_strdup(buffer);
                }
                // normal case, add the tag to the datalist
                else
                {
                    g_datalist_id_set_data_full(tags, g_quark_from_string(buffer_copy), g_strdup(buffer), g_free);
                }
                
                // the resolution is ready to be pushed to the datalist
                if (res_y!=NULL && res_x!=NULL)
                {
                    char* res = g_strdup_printf("%s x %s", res_x, res_y);
                    g_datalist_id_set_data_full(tags, g_quark_from_string(MEDIA_VIDEO_RESOLUTION_ATTRIBUTE), res, g_free);
                    res_x=NULL;
                    res_y=NULL;
                }
                
                // free the tag names
                g_free(buffer_copy);
                k=0;
            }
        }
        
        wait(pid);
        close(channel[0]);
        return tags;
    }
}


